@@ -13,7 +13,7 @@ import os
 import sys
 import csv as csv_std
 import tabulate
-from .utils import find_spatialite, sqlite3, decode_base64_values
+from .utils import file_progress, find_spatialite, sqlite3, decode_base64_values
 
 VALID_COLUMN_TYPES = ("INTEGER", "TEXT", "FLOAT", "BLOB")
 
@@ -45,11 +45,11 @@ def output_options(fn):
                 is_flag=True,
                 default=False,
             ),
-            click.option("-c", "--csv", is_flag=True, help="Output CSV"),
+            click.option("--csv", is_flag=True, help="Output CSV"),
+            click.option("--tsv", is_flag=True, help="Output TSV"),
             click.option("--no-headers", is_flag=True, help="Omit CSV headers"),
             click.option("-t", "--table", is_flag=True, help="Output as a table"),
             click.option(
-                "-f",
                 "--fmt",
                 help="Table format - one of {}".format(
                     ", ".join(tabulate.tabulate_formats)
@@ -120,6 +120,7 @@ def tables(
     nl,
     arrays,
     csv,
+    tsv,
     no_headers,
     table,
     fmt,
@@ -161,8 +162,8 @@ def tables(
 
     if table:
         print(tabulate.tabulate(_iter(), headers=headers, tablefmt=fmt))
-    elif csv:
-        writer = csv_std.writer(sys.stdout)
+    elif csv or tsv:
+        writer = csv_std.writer(sys.stdout, dialect="excel-tab" if tsv else "excel")
         if not no_headers:
             writer.writerow(headers)
         for row in _iter():
@@ -201,6 +202,7 @@ def views(
     nl,
     arrays,
     csv,
+    tsv,
     no_headers,
     table,
     fmt,
@@ -218,6 +220,7 @@ def views(
         nl=nl,
         arrays=arrays,
         csv=csv,
+        tsv=tsv,
         no_headers=no_headers,
         table=table,
         fmt=fmt,
@@ -584,6 +587,7 @@ def insert_upsert_options(fn):
                 help="Character encoding for input, defaults to utf-8",
             ),
             load_extension_option,
+            click.option("--silent", is_flag=True, help="Do not show progress bar"),
         )
     ):
         fn = decorator(fn)
@@ -608,6 +612,7 @@ def insert_upsert_implementation(
     default=None,
     encoding=None,
     load_extension=None,
+    silent=False,
 ):
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
@@ -621,9 +626,10 @@ def insert_upsert_implementation(
         pk = pk[0]
     if csv or tsv:
         dialect = "excel-tab" if tsv else "excel"
-        reader = csv_std.reader(json_file, dialect=dialect)
-        headers = next(reader)
-        docs = (dict(zip(headers, row)) for row in reader)
+        with file_progress(json_file, silent=silent) as json_file:
+            reader = csv_std.reader(json_file, dialect=dialect)
+            headers = next(reader)
+            docs = (dict(zip(headers, row)) for row in reader)
     elif nl:
         docs = (json.loads(line) for line in json_file)
     else:
@@ -673,6 +679,7 @@ def insert(
     alter,
     encoding,
     load_extension,
+    silent,
     ignore,
     replace,
     truncate,
@@ -702,6 +709,7 @@ def insert(
             truncate=truncate,
             encoding=encoding,
             load_extension=load_extension,
+            silent=silent,
             not_null=not_null,
             default=default,
         )
@@ -725,6 +733,7 @@ def upsert(
     default,
     encoding,
     load_extension,
+    silent,
 ):
     """
     Upsert records based on their primary key. Works like 'insert' but if
@@ -747,6 +756,7 @@ def upsert(
             default=default,
             encoding=encoding,
             load_extension=load_extension,
+            silent=silent,
         )
     except UnicodeDecodeError as ex:
         raise click.ClickException(UNICODE_ERROR.format(ex))
@@ -923,6 +933,7 @@ def query(
     nl,
     arrays,
     csv,
+    tsv,
     no_headers,
     table,
     fmt,
@@ -934,6 +945,7 @@ def query(
     "Execute SQL query and return the results as JSON"
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
+    db.register_fts4_bm25()
     with db.conn:
         cursor = db.execute(sql, dict(param))
         if cursor.description is None:
@@ -950,8 +962,8 @@ def query(
                 sys.stdout.write(str(data))
         elif table:
             print(tabulate.tabulate(list(cursor), headers=headers, tablefmt=fmt))
-        elif csv:
-            writer = csv_std.writer(sys.stdout)
+        elif csv or tsv:
+            writer = csv_std.writer(sys.stdout, dialect="excel-tab" if tsv else "excel")
             if not no_headers:
                 writer.writerow(headers)
             for row in cursor:
@@ -968,6 +980,87 @@ def query(
     required=True,
 )
 @click.argument("dbtable")
+@click.argument("q")
+@click.option("-o", "--order", type=str, help="Order by ('column' or 'column desc')")
+@click.option("-c", "--column", type=str, multiple=True, help="Columns to return")
+@click.option(
+    "--limit",
+    type=int,
+    help="Number of rows to return - defaults to everything",
+)
+@click.option(
+    "--sql", "show_sql", is_flag=True, help="Show SQL query that would be run"
+)
+@output_options
+@load_extension_option
+@click.pass_context
+def search(
+    ctx,
+    path,
+    dbtable,
+    q,
+    order,
+    show_sql,
+    column,
+    limit,
+    nl,
+    arrays,
+    csv,
+    tsv,
+    no_headers,
+    table,
+    fmt,
+    json_cols,
+    load_extension,
+):
+    "Execute a full-text search against this table"
+    db = sqlite_utils.Database(path)
+    _load_extensions(db, load_extension)
+    # Check table exists
+    table_obj = db[dbtable]
+    if not table_obj.exists():
+        raise click.ClickException("Table '{}' does not exist".format(dbtable))
+    if not table_obj.detect_fts():
+        raise click.ClickException(
+            "Table '{}' is not configured for full-text search".format(dbtable)
+        )
+    if column:
+        # Check they all exist
+        table_columns = table_obj.columns_dict
+        for c in column:
+            if c not in table_columns:
+                raise click.ClickException(
+                    "Table '{}' has no column '{}".format(dbtable, c)
+                )
+    sql = table_obj.search_sql(columns=column, order_by=order, limit=limit)
+    if show_sql:
+        click.echo(sql)
+        return
+    ctx.invoke(
+        query,
+        path=path,
+        sql=sql,
+        nl=nl,
+        arrays=arrays,
+        csv=csv,
+        tsv=tsv,
+        no_headers=no_headers,
+        table=table,
+        fmt=fmt,
+        json_cols=json_cols,
+        param=[("query", q)],
+        load_extension=load_extension,
+    )
+
+
+@cli.command()
+@click.argument(
+    "path",
+    type=click.Path(file_okay=True, dir_okay=False, allow_dash=False),
+    required=True,
+)
+@click.argument("dbtable")
+@click.option("-c", "--column", type=str, multiple=True, help="Columns to return")
 @output_options
 @load_extension_option
 @click.pass_context
@@ -975,9 +1068,11 @@ def rows(
     ctx,
     path,
     dbtable,
+    column,
     nl,
     arrays,
     csv,
+    tsv,
     no_headers,
     table,
     fmt,
@@ -985,13 +1080,17 @@ def rows(
     load_extension,
 ):
     "Output all rows in the specified table"
+    columns = "*"
+    if column:
+        columns = ", ".join("[{}]".format(c) for c in column)
     ctx.invoke(
         query,
         path=path,
-        sql="select * from [{}]".format(dbtable),
+        sql="select {} from [{}]".format(columns, dbtable),
         nl=nl,
         arrays=arrays,
         csv=csv,
+        tsv=tsv,
         no_headers=no_headers,
         table=table,
         fmt=fmt,
