@@ -1,5 +1,6 @@
 from .utils import sqlite3, OperationalError, suggest_column_types, column_affinity
 from collections import namedtuple, OrderedDict
+from collections.abc import Mapping
 import contextlib
 import datetime
 import decimal
@@ -9,11 +10,33 @@ import itertools
 import json
 import os
 import pathlib
+import re
+from sqlite_fts4 import rank_bm25
+import sys
 import textwrap
 import uuid
 
 SQLITE_MAX_VARS = 999
 
+_virtual_table_using_re = re.compile(
+    r"""
+^ # Start of string
+\s*CREATE\s+VIRTUAL\s+TABLE\s+ # CREATE VIRTUAL TABLE
+(
+    '(?P<squoted_table>[^']*(?:''[^']*)*)' | # single quoted name
+    "(?P<dquoted_table>[^"]*(?:""[^"]*)*)" | # double quoted name
+    `(?P<backtick_table>[^`]+)`            | # `backtick` quoted name
+    \[(?P<squarequoted_table>[^\]]+)\]     | # [...] quoted name
+    (?P<identifier>                          # SQLite non-quoted identifier
+        [A-Za-z_\u0080-\uffff]  # \u0080-\uffff = "any character larger than u007f"
+        [A-Za-z_\u0080-\uffff0-9\$]* # zero-or-more alphanemuric or $
+    )
+)
+\s+(IF\s+NOT\s+EXISTS\s+)?      # IF NOT EXISTS (optional)
+USING\s+(?P<using>\w+)          # e.g. USING FTS5
+""",
+    re.VERBOSE | re.IGNORECASE,
+)
 
 try:
     import pandas as pd
@@ -132,6 +155,7 @@ class Database:
         self._tracer = tracer
         if recursive_triggers:
             self.execute("PRAGMA recursive_triggers=on;")
+        self._registered_functions = set()
 
     @contextlib.contextmanager
     def tracer(self, tracer=None):
@@ -148,10 +172,26 @@ class Database:
     def __repr__(self):
         return "<Database {}>".format(self.conn)
 
-    def register_function(self, fn):
-        name = fn.__name__
-        arity = len(inspect.signature(fn).parameters)
-        self.conn.create_function(name, arity, fn)
+    def register_function(self, fn=None, deterministic=None, replace=False):
+        def register(fn):
+            name = fn.__name__
+            arity = len(inspect.signature(fn).parameters)
+            if not replace and (name, arity) in self._registered_functions:
+                return fn
+            kwargs = {}
+            if deterministic and sys.version_info >= (3, 8):
+                kwargs["deterministic"] = True
+            self.conn.create_function(name, arity, fn, **kwargs)
+            self._registered_functions.add((name, arity))
+            return fn
+
+        if fn is None:
+            return register
+        else:
+            register(fn)
+
+    def register_fts4_bm25(self):
+        self.register_function(rank_bm25, deterministic=True)
 
     def execute(self, sql, parameters=None):
         if self._tracer:
@@ -664,6 +704,14 @@ class Table(Queryable):
                 )
         return fks
 
+    @property
+    def virtual_table_using(self):
+        "Returns type of virtual table or None if this is not a virtual table"
+        match = _virtual_table_using_re.match(self.schema)
+        if match is None:
+            return None
+        return match.groupdict()["using"].upper()
+
     @property
     def indexes(self):
         sql = 'PRAGMA index_list("{}")'.format(self.name)
@@ -1273,21 +1321,70 @@ class Table(Queryable):
             )
         return self
 
-    def search(self, q):
-        sql = (
-            textwrap.dedent(
-                """
-            select * from "{table}" where rowid in (
-                select rowid from [{table}_fts]
-                where [{table}_fts] match :search
+    def search_sql(self, columns=None, order_by=None, limit=None):
+        # Pick names for table and rank column that don't clash
+        original = "original_" if self.name == "original" else "original"
+        columns_sql = "*"
+        columns_with_prefix_sql = "[{}].*".format(original)
+        if columns:
+            columns_sql = ",\n        ".join("[{}]".format(c) for c in columns)
+            columns_with_prefix_sql = ",\n    ".join(
+                "[{}].[{}]".format(original, c) for c in columns
             )
-            order by rowid
-        """
-            )
-            .strip()
-            .format(table=self.name)
+        fts_table = self.detect_fts()
+        assert fts_table, "Full-text search is not configured for table '{}'".format(
+            self.name
         )
-        return self.db.execute(sql, (q,)).fetchall()
+        virtual_table_using = self.db[fts_table].virtual_table_using
+        sql = textwrap.dedent(
+            """
+        with {original} as (
+            select
+                rowid,
+                {columns}
+            from [{dbtable}]
+        )
+        select
+            {columns_with_prefix}
+        from
+            [{original}]
+            join [{fts_table}] on [{original}].rowid = [{fts_table}].rowid
+        where
+            [{fts_table}] match :query
+        order by
+            {order_by}
+        {limit}
+        """
+        ).strip()
+        if virtual_table_using == "FTS5":
+            rank_implementation = "[{}].rank".format(fts_table)
+        else:
+            self.db.register_fts4_bm25()
+            rank_implementation = "rank_bm25(matchinfo([{}], 'pcnalx'))".format(
+                fts_table
+            )
+        return sql.format(
+            dbtable=self.name,
+            original=original,
+            columns=columns_sql,
+            columns_with_prefix=columns_with_prefix_sql,
+            fts_table=fts_table,
+            order_by=order_by or rank_implementation,
+            limit="limit {}".format(limit) if limit else "",
+        ).strip()
+
+    def search(self, q, order_by=None, columns=None, limit=None):
+        cursor = self.db.execute(
+            self.search_sql(
+                order_by=order_by,
+                columns=columns,
+                limit=limit,
+            ),
+            {"query": q},
+        )
+        columns = [c[0] for c in cursor.description]
+        for row in cursor:
+            yield dict(zip(columns, row))
 
     def value_or_default(self, key, value):
         return self._defaults[key] if value is DEFAULT else value
@@ -1772,15 +1869,20 @@ class Table(Queryable):
             return pk
 
     def m2m(
-        self, other_table, record_or_list=None, pk=DEFAULT, lookup=None, m2m_table=None
+        self,
+        other_table,
+        record_or_iterable=None,
+        pk=DEFAULT,
+        lookup=None,
+        m2m_table=None,
     ):
         if isinstance(other_table, str):
             other_table = self.db.table(other_table, pk=pk)
         our_id = self.last_pk
         if lookup is not None:
-            assert record_or_list is None, "Provide lookup= or record, not both"
+            assert record_or_iterable is None, "Provide lookup= or record, not both"
         else:
-            assert record_or_list is not None, "Provide lookup= or record, not both"
+            assert record_or_iterable is not None, "Provide lookup= or record, not both"
         tables = list(sorted([self.name, other_table.name]))
         columns = ["{}_id".format(t) for t in tables]
         if m2m_table is not None:
@@ -1801,10 +1903,11 @@ class Table(Queryable):
                 m2m_table_name = m2m_table or "{}_{}".format(*tables)
         m2m_table = self.db.table(m2m_table_name, pk=columns, foreign_keys=columns)
         if lookup is None:
+            # if records is only one record, put the record in a list
             records = (
-                [record_or_list]
-                if not isinstance(record_or_list, (list, tuple))
-                else record_or_list
+                [record_or_iterable]
+                if isinstance(record_or_iterable, Mapping)
+                else record_or_iterable
             )
             # Ensure each record exists in other table
             for record in records:
